# encoding: utf-8
# frozen_string_literal: true

class UTAUSequence

    Note = Struct.new :length,:lyric,:notenum,:intensity,:modulation do
        def self.from_ust(section)
            note = self.new
            section.lines do |line|
                raise "Can't parse #{line.inspect}" unless line =~ /^(\w+)=(.*)$/
                setter = :"#{$1.downcase}="
                case setter
                when :lyric=
                    note.lyric = $2
                when :tempo=,:preutterance=,:flags=,:velocity=,:pbs=,:pbw=,:pby=,:pbm=
                    # Let's ignore this
                when :pbtype=,:pitchbend=,:pbstart=,:envelope=,:voiceoverlap=,:vbr=
                    puts "TODO: ignoring #{$1} record"
                else
                    note.send setter,Integer($2)
                end
            end
            return note
        end

        def rest?
            !!(lyric =~ /\AR\Z/i)
        end

        #def hz
        #    2.0**((notenum-69)/12.0) * 440.0
        #end

        def tract_gp
            notenum*4 - 128
        end
    end

    attr_accessor :notes

    def initialize(notes=[])
        @notes = notes
    end

    def self.load_file(path)
        self.new(File.read(path).scan(/\[#(\w+)\]\n(.*?)(?=(\[|\z))/m).map{|(section,data,_)| section=~/\A\d+\Z/ ? Note.from_ust(data) : nil}.compact)
    end

end

class TractSequence

    PARAMETERS = %i[aa ga gp vp vr f1 f2 f3 f4 na nf fa ff].freeze

    attr_accessor :points

    TractPoint = Struct.new :length,*PARAMETERS do

        def encode_vlv(num)
            if num < 0x80
                "byte #{num}"
            elsif num < 0x8000
                "byte #{num>>8}+128,#{num&255}"
            else
                raise "#{num} too big for VLV encoding"
            end
        end

        def to_spin(last)
            bitmask = 0
            pstr = String.new
            PARAMETERS.each_with_index do |par,i|
                self[par] = last[par]||0 if self[par].nil?
                if self[par] != last[par]
                    pstr << "byte $#{"%02X"%self[par]} '#{par}\n"
                    bitmask |= 1<<i
                end
            end

            return "#{encode_vlv length}\n#{encode_vlv bitmask}\n#{pstr}\n"
        end
    end

    def initialize
        @points = []
    end

    def to_spin
        data = points.reduce([String.new,TractPoint.new]){|(s,last),now|[s + now.to_spin(last),now]}[0]
        <<~SPINHEAD
        PUB get_track
          return @vtrack
        DAT
        vtrack
        #{data}

        byte 0
        byte word $FFFF ' end of track
        SPINHEAD
    end

end

class WorkSequence
    
    PARAMETERS = TractSequence::PARAMETERS

    attr_accessor :points
    attr_accessor :level

    def initialize(level)
        @points = PARAMETERS.map{|par|[par,{}]}.to_h
        @level = level
    end

    def self.from_utau(utau,tempomul,voicemul,level,transpose)
        seq = self.new(level)
        u2w = UTAU2Work.new(seq,tempomul,voicemul,transpose)
        u2w.do_it(utau)
        return seq
    end

    def par2byte(par,val)
        byte = case par
        when :aa,:ga,:fa
            (val*level*255).round
        when :gp,:vp
            val.round
        when :f1,:f2,:f3,:f4,:nf
            (val/19.53).round
        when :na
            (val*255).round
        when :ff
            (val/39.06).round
        when :vr
            (val/0.0763).round
        else raise
        end
        raise "value #{val.inspect} for parameter #{par} went out of range (#{byte} not in 0..255)" unless (0..255).include? byte
        return byte
    end

    def to_tract()
        seq = TractSequence.new
        lastvals = {}
        lastparpos  = PARAMETERS.map{|par|[par,-1]}.to_h
        parposstacks = PARAMETERS.map{|par|[par,points[par].keys.sort!.reverse!]}.to_h
        exportpos = 0

        loop do
            lastpos = exportpos
            exportpos = parposstacks.values.map(&:last).compact.min
            #p exportpos
            
            seq.points << TractSequence::TractPoint.new(exportpos-lastpos,
                *PARAMETERS.map do |par| # Interpolate all values
                    prevt = lastparpos[par]
                    nextt = parposstacks[par].last
                    prevv = lastvals[par]
                    nextv = points[par][nextt]
                    if nextv.nil?
                        next nil
                    elsif prevv.nil?
                        next par2byte(par,nextv)
                    else
                        interp = nextt == prevt ? 1 : (exportpos - prevt).to_f/(nextt-prevt).to_f
                        val = nextv*interp + prevv*(1-interp)
                        next par2byte(par,val)
                    end
                end
            )

            parposstacks.each do |par,stack|
                until stack.last.nil? || stack.last > exportpos
                    pos = stack.pop
                    lastvals[par] = points[par][pos]
                    lastparpos[par] = exportpos
                end
            end

            break if parposstacks.values.all? &:empty?
        end
        return seq
    end

    class UTAU2Work

        attr_reader   :seq,:tempomul,:voicemul,:transpose
        attr_accessor :utaupos,:note

        def initialize(seq,tempomul,voicemul,transpose)
            @seq = seq
            @tempomul = tempomul
            @voicemul = voicemul
            @utaupos = 0
            @transpose = transpose
        end

        # Push a control point...
        def pushcp(reltime,**pars)
            abstime = :oh_no
            case reltime
            when 0...1
                abstime = utaupos + @note.length*reltime
            when 1..
                abstime = utaupos + reltime*voicemul
            when -Float::INFINITY...0
                abstime = utaupos + @note.length - reltime*voicemul
                abstime = utaupos if abstime < utaupos # Push control point forward to fit
            else raise "Can't handle reltime #{reltime}"
            end
            #p abstime
            ttime = (abstime*tempomul).to_i
            pars.each_pair do |par,v|
                seq.points[par][ttime] = v.to_f
            end
        end

        FORMANT_TBL = {
=begin

            'a' => {f1:750,f2:1200,f3:2500,f4:4000},
            'e' => {f1:360,f2:1400,f3:2160,f4:3200},
            'i' => {f1:400,f2:2000,f3:2550,f4:3900}, 
            'o' => {f1:380,f2:870,f3:2250,f4:3900}, 
            'u' => {f1:470,f2:1120,f3:2430,f4:3400}, 
=end
            'a' => {f1:750,f2:1200,f3:2500,f4:4000},
            'e' => {f1:360,f2:1400,f3:2160,f4:3200},
            'i' => {f1:400,f2:2000,f3:2550,f4:3900}, 
            'o' => {f1:380,f2:870,f3:2250,f4:3900}, 
            'u' => {f1:470,f2:1120,f3:2430,f4:3400}, 

            'n' => {f1:610,f2:1640,f3:2460,f4:4000,nf:1900},
            'm' => {f1:300,f2:1400,f3:2160,f4:3200,nf:274},
            'w' => {f1:300,f2:1400,f3:2160,f4:3200,nf:1900},

            'r' => {f1:360,f2:1060,f3:1380,f4:3500}, #???

            'y' => {f1:380,f2:1760,f3:3680,f4:4500}, #???

            'b' => {f1:100,f2:1100,f3:2150,f4:3500},

            's' => {f1:50,f2:1060,f3:1380,f4:3500},
        }

        def getformants(phon,note)
            fshift = (2.0**((note-148)/48.0) -1)*0.1+1
            raise "Unnown FORMANT_TBL entry #{phon.inspect}" unless FORMANT_TBL.key? phon
            return FORMANT_TBL[phon].transform_values{|f|f*fshift}
        end

        VOCAL_FENVELOPE = {
            'a' => +1,
            'e' => 0,
            'i' => -1,
            'o' => 0,
            'u' => -1,
        }


        def do_it(utau)
            gpnote = nil
            utau.notes.each do |__note|
                @note = __note

                next if note.length == 0

                begin
                    if @note.rest?
                        pushcp 200,aa:0.0,ga:0.0,fa:0.0,gp:gpnote,vp:0
                        if note.length >= 5000
                            pushcp 2100,aa:0.0,gp:gpnote
                            pushcp 2400,aa:0.3
                            pushcp 2700,aa:0.0

                            pushcp 3100,aa:0.0,gp:gpnote
                            pushcp 4200,aa:0.3
                            pushcp 4700,aa:0.0
                        end
                        pushcp -1,aa:0.0,ga:0.0,fa:0.0,vp:0
                    else
                        gpnote = note.tract_gp + transpose
                        enn = false

                        lyric = note.lyric
                        #lyric = 'ya'
                        case lyric

                        when 'ka','ke','ki','ko','ku','ga','ge','gi','go','gu' # Ka: ok, ga: terrible
                            vocal = lyric[-1]
                            is_k = lyric.start_with? 'k'
                            #p [lyric,vocal,is_k]
                            vf = getformants(vocal,gpnote)
                            kf = is_k ? 
                                {f1:vf[:f1]*0.75, f2:vf[:f2]*0.75+vf[:f3]*0.25, f3:vf[:f2]*0.25+vf[:f3]*0.75, f4:vf[:f4]} :
                                {f1:vf[:f1]*0.25, f2:vf[:f2]*0.5+vf[:f3]*0.5, f3:vf[:f3], f4:vf[:f4]}

                            pushcp  0,gp:gpnote,aa:0,**kf,na:0,aa:0,ga:0,ff:90,fa:0
                            pushcp 25,ga:0.6,gp:gpnote-3,**kf,aa:0,fa:(is_k ? 0.1 : 0.02)
                            pushcp 50,ga:0.1,gp:gpnote,**kf,aa:0.8,ff:90,fa:0

                            pushcp 200,ga:0.9,**vf,aa:0,ff:90
        
                            pushcp 0.85,ga:0.85,gp:gpnote+VOCAL_FENVELOPE[vocal],**vf,na:0,fa:0
                            pushcp 0.95,ga:0.55

                        when 'wa','we','wi','wo','wu' # kinda eh

                            vocal = lyric[-1]
                            pushcp  0,gp:gpnote,aa:0,**getformants('w',gpnote),na:0.6,fa:0,ff:200
                            pushcp 0.1,ga:0.6,gp:gpnote,fa:0.2

                            pushcp 0.33,**getformants('w',gpnote)

                            pushcp 0.4,**getformants(vocal,gpnote),na:0,ga:0.9,fa:0,fa:0.1
        
                            pushcp 0.85,ga:0.9,gp:gpnote,**getformants(vocal,gpnote),na:0,fa:0,ff:200
                            pushcp 0.95,ga:0.6

                        when 'fa','fe','fi','fo','fu' # terrible
                            vocal = lyric[-1]
                            
                            pushcp  0,gp:gpnote,aa:0,na:0,ff:1000,**getformants('s',gpnote)
                            pushcp 0.1,ga:0.0,gp:gpnote,fa:0.3,aa:0.2,**getformants(vocal,gpnote)
                            pushcp 0.3,ga:0.1,aa:00,ff:980,gp:gpnote

                            pushcp 0.3,ga:0.8,fa:0.2,aa:00,gp:gpnote
        
                            pushcp 0.85,ga:0.75,gp:gpnote+VOCAL_FENVELOPE[vocal],**getformants(vocal,gpnote),na:0,fa:0
                            pushcp 0.95,ga:0.55

                        when 'ta','te','ti','to','tu','da','de','di','do','du' # terrible
                            vocal = lyric[-1]
                            
                            pushcp  0,gp:gpnote,aa:0,na:0,ff:3800,**getformants(vocal,gpnote)
                            pushcp  25,gp:gpnote,aa:1.0,na:0,ff:3800
                            pushcp 50,ga:0.0,gp:gpnote,fa:0.3,aa:0.1,**getformants(vocal,gpnote)
                            pushcp 120,ga:0.1,fa:0,aa:00,ff:3800,gp:gpnote

                            pushcp 0.4,ga:0.9,fa:0,aa:00,gp:gpnote
        
                            pushcp 0.85,ga:0.75,gp:gpnote+VOCAL_FENVELOPE[vocal],**getformants(vocal,gpnote),na:0,fa:0
                            pushcp 0.95,ga:0.55


                        when 'tsa','tse','tsi','tso','tsu' # actually kinda okay
                            vocal = lyric[-1]
                            
                            pushcp  0,gp:gpnote,aa:0,na:0,ff:3800,**getformants(vocal,gpnote)
                            pushcp  0.1,gp:gpnote,aa:1.0,na:0,ff:3800,**getformants('s',gpnote)
                            pushcp 0.2,ga:0.0,gp:gpnote,fa:0.3,aa:0.1,**getformants(vocal,gpnote)
                            pushcp 0.4,ga:0.1,fa:0,aa:00,ff:3800,gp:gpnote

                            pushcp 0.4,ga:0.9,fa:0,aa:00,gp:gpnote
        
                            pushcp 0.85,ga:0.75,gp:gpnote+VOCAL_FENVELOPE[vocal],**getformants(vocal,gpnote),na:0,fa:0
                            pushcp 0.95,ga:0.55

                        when 'sa','se','si','so','su','za','ze','zi','zo','zu' # terrible
                            vocal = lyric[-1]
                            
                            pushcp  0,gp:gpnote,aa:0,na:0,ff:3800,**getformants('s',gpnote)
                            pushcp 50,ga:0.0,gp:gpnote,fa:0.3,aa:0.2,**getformants(vocal,gpnote)
                            pushcp 150,ga:0.1,fa:0,aa:00,ff:3800,gp:gpnote

                            pushcp 0.3,ga:0.9,fa:0,aa:00,gp:gpnote
        
                            pushcp 0.85,ga:0.9,gp:gpnote+VOCAL_FENVELOPE[vocal],**getformants(vocal,gpnote),na:0,fa:0
                            pushcp 0.95,ga:0.6

                        when 'sha','she','shi','sho','shu','cha','che','chi','cho','chu' # terrible
                            vocal = lyric[-1]
                            
                            pushcp  0,gp:gpnote,aa:0,**getformants(vocal,gpnote),na:0,aa:0,fa:0,ff:1900
                            pushcp 0.16,ga:0,**getformants(vocal,gpnote),fa:0.7,aa:0.1
                            pushcp 0.3,ga:0.2,**getformants(vocal,gpnote),aa:0.8,gp:gpnote

                            pushcp 0.5,ga:0.9,**getformants(vocal,gpnote),aa:0.3,fa:0,ff:2000
        
                            pushcp 0.85,ga:0.9,gp:gpnote+VOCAL_FENVELOPE[vocal],**getformants(vocal,gpnote),na:0,aa:0,fa:0
                            pushcp 0.95,ga:0.6

                        when 'ba','be','bi','bo','bu' # kinda terrible
                            vocal = lyric[-1]
                            
                            pushcp  0,gp:gpnote,aa:0,**getformants('b',gpnote),na:0,aa:0
                            pushcp 0.16,ga:0.9,gp:gpnote-1,**getformants('b',gpnote),aa:0
                            pushcp 0.28,ga:0.5,gp:gpnote-1,**getformants('b',gpnote),aa:0.8

                            pushcp 0.45,ga:0.85,**getformants(vocal,gpnote),aa:0
        
                            pushcp 0.85,ga:0.85,gp:gpnote+VOCAL_FENVELOPE[vocal],**getformants(vocal,gpnote),na:0,fa:0
                            pushcp 0.95,ga:0.55

                        when 'ra','re','ri','ro','ru' # pretty ok
                            vocal = lyric[-1]
                            long = note.length >= 1000
                            
                            pushcp  0,gp:gpnote-1,aa:0,**getformants('r',gpnote),na:0

                        
                            pushcp (long ? 100 : 0.1),ga:0.6,gp:gpnote-1,**getformants('r',gpnote)

                            pushcp (long ? 240 : 0.24),ga:0.6,gp:gpnote
                            pushcp (long ? 490 : 0.54),ga:0.85,gp:gpnote,**getformants(vocal,gpnote)
        
                            pushcp 0.85,ga:0.85,gp:gpnote+VOCAL_FENVELOPE[vocal],**getformants(vocal,gpnote),na:0,fa:0
                            pushcp 0.95,ga:0.55

                        when 'ha','he','hi','ho','hu' # good
                            vocal = lyric[-1]
                            pushcp  0,gp:gpnote,aa:0,na:0,**getformants(vocal,gpnote)
                            pushcp 0.2,ga:0.1,aa:0.8,gp:gpnote

                            pushcp 0.55,ga:1.0
        
                            pushcp 0.85,ga:0.75,gp:gpnote+VOCAL_FENVELOPE[vocal],**getformants(vocal,gpnote),na:0,fa:0
                            pushcp 0.95,ga:0.55

                        when 'ma','me','mi','mo','mu' # as good as i can get it
                            vocal = lyric[-1]
                            pushcp  0,gp:gpnote-1,aa:0,**getformants('m',gpnote),na:1.0
                            pushcp 0.1,ga:0.7,gp:gpnote-1,**getformants('m',gpnote)

                            pushcp 0.3,**getformants('m',gpnote),na:1.0

                            pushcp 0.45,**getformants(vocal,gpnote),na:0.35,ga:0.9,gp:gpnote
        
                            pushcp 0.85,ga:0.9,gp:gpnote+VOCAL_FENVELOPE[vocal],**getformants(vocal,gpnote),na:0,fa:0
                            pushcp 0.95,ga:0.6

                        when 'na','ne','ni','no','nu' # as good as i can get it
                            vocal = lyric[-1]
                            pushcp  0,gp:gpnote,aa:0,**getformants('n',gpnote),na:1.0
                            pushcp 0.1,ga:0.7,gp:gpnote

                            pushcp 0.55,**getformants(vocal,gpnote),na:0.35,ga:0.9
        
                            pushcp 0.85,ga:0.9,gp:gpnote+VOCAL_FENVELOPE[vocal],**getformants(vocal,gpnote),na:0,fa:0
                            pushcp 0.95,ga:0.6

                        when 'ya','ye','yi','yo','yu','ja','je','ji','jo','ju' # ok
                            vocal = lyric[-1]
                            pushcp  0,gp:gpnote-1,aa:0,**getformants('y',gpnote),na:0.5
                            pushcp 0.1,ga:0.7,gp:gpnote-1,**getformants('y',gpnote)

                            pushcp 0.3,**getformants('y',gpnote),na:0.5

                            pushcp 0.45,**getformants(vocal,gpnote),na:0.15,ga:0.9,gp:gpnote
        
                            pushcp 0.85,ga:0.9,gp:gpnote+VOCAL_FENVELOPE[vocal],**getformants(vocal,gpnote),na:0,fa:0
                            pushcp 0.95,ga:0.6

                        when 'a','e','i','o','u' # kinda bad
                            vocal = lyric
                            pushcp  0,gp:gpnote,aa:0,**getformants(vocal,gpnote),na:0
                            pushcp 0.1,ga:1.0,gp:gpnote
                            pushcp 0.3,ga:0.75
        
                            pushcp 0.85,ga:0.75,gp:gpnote+VOCAL_FENVELOPE[vocal],**getformants(vocal,gpnote),na:0,fa:0
                            pushcp 0.95,ga:0.55
                        when 'n','nn' # should test agian
                            pushcp  0,gp:gpnote,aa:0,na:1.0,**getformants('n',gpnote)
                            pushcp 0.1,ga:1.0,gp:gpnote,**getformants('n',gpnote)

                            pushcp 0.85,ga:0.75,gp:gpnote,**getformants('n',gpnote),na:1.0,fa:0
                            pushcp 0.95,ga:0.55
                        else # fallback
                            puts "fallback for #{lyric}"
                            vocal = lyric[-1]
                            pushcp  0,gp:gpnote,aa:0,**getformants(vocal,gpnote),na:0
                            pushcp 0.1,ga:1.0,gp:gpnote,**getformants(vocal,gpnote)
                            pushcp 0.3,ga:0.75
        
                            pushcp 0.85,ga:0.75,gp:gpnote+VOCAL_FENVELOPE[vocal],**getformants(vocal,gpnote),na:0,fa:0
                            pushcp 0.95,ga:0.55
                        end
                    end
                rescue 
                    puts "exception occured while processing #{note.inspect}"
                    raise
                end

                @utaupos += @note.length
            end
            note = nil
        end
    end
end

if $0 == __FILE__
    #utau = UTAUSequence.load_file 'aft55.ust'
    #work = WorkSequence.from_utau utau,1/1r,1,0.1,-24

    utau = UTAUSequence.load_file 'lagtrain.ust'
    work = WorkSequence.from_utau utau,1.213,1,0.1,0

    #utau = UTAUSequence.load_file 'after.ust'
    #work = WorkSequence.from_utau utau,1/1r,1,0.1,0

    #utau = UTAUSequence.load_file 'meltdown_mini.ust'
    #work = WorkSequence.from_utau utau,1/1r,1,0.1,-48

    #utau = UTAUSequence.load_file 'karakuri_rin.ust'
    #work = WorkSequence.from_utau utau,1.0,1,0.1,0

    #utau = UTAUSequence.load_file '39music.ust'
    #work = WorkSequence.from_utau utau,2,1,0.1,0

    tract = work.to_tract
    File.write 'lagtrain.spin',tract.to_spin
end
